// icat.spat tools by Ivica Ico Bukvic (c) 2014 <ico@vt.edu>
// distributed under GNU GPL license 3 or greater
// for more info on the license visit http://www.gnu.org/copyleft/gpl.html

import com.cycling74.max.*;

public class icat_spat_vertical extends MaxObject {

	
	
  private float[] layers;
  private float l_angle = 0;
  private float h_angle = 0;
  private float gain = 1;
  private float range = 180;
  private float h_range = 360;
  private int length = 0;

  
  
  public icat_spat_vertical()
  {
	declareIO(1,1);
	setInletAssist(new String[] { "float list or setup" });
	setOutletAssist(new String[] { "float list: layer, layer angle, layer amplitude, gain" });
  }

  
  
  //main calculation of the amplitude ratios
  public void inlet(float val)
  {
  	if (length == 0)
  	{
  		error("icat_spat_vertical: no speaker setup provided\n");
  		return;
  	}
  	
  	int lower = -1;
  	int higher = -1;
  	float l_amp = 0;
  	float h_amp = 0;
  	int i = 0;
  	float distance = 0;
  	
  	//post("===============================\n");
  	
  	if (val < 0)
  		while (val < 0)
  			val += range * 2;
  	else if (val > range * 2)
  		while (val > range * 2)
  			val -= range * 2;

  	for (i = 0; i < length; i++)
  	{
  		if (layers[i] <= val)
  		{
  			//post("lower=" + lower + " i=" + i + "\n");
  			lower = i;
  		}
  		else
  		{
  			//post("higher=" + higher + " i=" + i + "\n");
  			higher = i;
  			break;
  		}
  	}
  	
  	if (lower == -1)
  		lower = length - 1;
  	if (higher == -1)
  		higher = lower + 1;
  	// this will make right translate sound across the floor in case we are dealing with a hemisphere
  	if (higher >= length)
  		higher -= length;

  	//post("lower=" + lower + " higher=" + higher + " l_angle=h_angle=" + l_angle + " val=" + val + "\n");

  	// if we are over the half of the angle (for the CUBE) we are mirroring on the other side	
  	if (lower >= (length / 2))
  	{
  		//post("l_angle flip\n");
  		l_angle += 180;
  		while (l_angle > h_range)
  			l_angle -= h_range;
  	}
  	if (higher >= (length / 2))
  	{
  		//post("h_angle flip\n");
  		h_angle += 180;
  		while (h_angle > h_range)
  			h_angle -= h_range;		
  	}
  	
  	//post("lower=" + lower + " higher=" + higher + " l_angle=" + l_angle + " h_angle=" + h_angle + "\n");
  	
  	if (higher != 0)
  		distance = layers[higher] - layers[lower];
  	else
  		distance = (range * 2) + layers[higher] - layers[lower];
  		
  	distance = Math.abs(distance);
  	//if (distance == 0)
  	//	distance = ((range / 2) - layers[length - 1]) * 2;
  	//post("distance=" + distance + "\n");
  	
  	if (distance != 0)
  	{
  	  	// we add range (360) in cases where val is e.g. 0 and is in between two speakers, neither of which are at 0
  	  	if (val < layers[lower])
  	  		val += range * 2;
  		l_amp = (float)Math.sqrt(1 - (val - layers[lower])/distance);
  		h_amp = (float)Math.sqrt((val - layers[lower])/distance);
  	}
  	
  	//post("l_amp=" + l_amp + " h_amp=" + h_amp + " length=" + length + "\n");
  	
  	//output speaker levels including their inverted angle (masked) component
  	int mask = 0;
  	
  	for (i = 0; i < length; i++)
  	{
  		if (i >= length / 2)
  			mask = 1;
		if (i == lower)
  			outlet(0, new Atom[]{Atom.newAtom(i), Atom.newAtom(l_angle), Atom.newAtom(l_amp), Atom.newAtom(gain), Atom.newAtom(mask)});
  		else if (i == higher)
  			outlet(0, new Atom[]{Atom.newAtom(i), Atom.newAtom(h_angle), Atom.newAtom(h_amp), Atom.newAtom(gain), Atom.newAtom(mask)});
  		else
  			outlet(0, new Atom[]{Atom.newAtom(i), Atom.newAtom(0), Atom.newAtom(0), Atom.newAtom(0), Atom.newAtom(mask)});
  	}
  	//now output sub level
  	outlet(0, new Atom[]{Atom.newAtom("sub"), Atom.newAtom(l_angle), Atom.newAtom(l_amp), Atom.newAtom(gain), Atom.newAtom(0)});
  	outlet(0, new Atom[]{Atom.newAtom("sub"), Atom.newAtom(h_angle), Atom.newAtom(h_amp), Atom.newAtom(gain), Atom.newAtom(mask)});
  }
  
  
  
  //redirect for integer values to the main calculation function above
  public void inlet(int val)
  {
	inlet((float)val);
  }

  
  
  //set up speaker array layers on a vertical plane
  public void setup(float[] arguments)
  {
	if (arguments.length > 0)
	{
		int i = 0;
		layers = null;
		layers = new float[arguments.length * 2]; // we add mirrored ones
		
		for (i = 0; i < arguments.length; i++)
		{
			if (arguments[i] < 0 || arguments[i] >= range)
				error("icat_spat_vertical: speaker array should have angles between 0 and " + range + "\n");
			else if (i > 0 && arguments[i] < arguments[i-1])
				error("icat_spat_vertical: speaker array should be provided sequentially in ascending order from 0 to "+ range + "\n");
			else
			{
				if (arguments[i] >= range/2)
					error("icat_spat_vertical: speaker angles need to be less than half of the range (default 180)\n");
				else
				{
					layers[i] = arguments[i];
					//post("speaker " + i + " " + arguments[i] + "\n");
				}
			}
		}
		// now we add mirrored ones
		for (i = 0; i < arguments.length; i++)
		{
			layers[arguments.length + i] = (range / 2) + ((range / 2) - arguments[arguments.length - 1 - i]);
			//post("speaker " + (arguments.length + i) + " " + layers[arguments.length + i] + "\n");
		}
		length = layers.length;
	}
  }

  
  
  //list of three values -- default implementation when using with ico.sound~ and ico.spat.horizontal script
  public void list(Atom[] arguments)
  {
	if (arguments.length == 3)
	{
		int i = 0;
		for (i = 2; i >= 0; i--)
		{
			if (arguments[i].isFloat())
			{
				switch (i)
				{
					case 0:
						inlet(arguments[i].getFloat());
						break;
					case 1:
						l_angle = arguments[i].getFloat();
						h_angle = arguments[i].getFloat();
						break;
					case 2:
						gain = arguments[i].getFloat();
						break;
					default:
						break;
				}
					
			}
			else if (arguments[i].isInt())
			{
				switch (i)
				{
					case 0:
						inlet(arguments[i].getInt());
						break;
					case 1:
						l_angle = arguments[i].getInt();
						h_angle = arguments[i].getInt();
						break;
					case 2:
						gain = arguments[i].getInt();
						break;
					default:
						break;
				}
			}
			else 
			{
				error("icat_spat_vertical: invalid values (should be int/float list)\n");
				break;
			}
		}
	}
  }
 
}
